@@ -1,5 +1,6 @@
 from ossapi.utils import BaseModel
 
+# fmt: off
 int_to_mod = {
     0          : ["NM",       "NoMod"],
     1 << 0     : ["NF",      "NoFail"],
@@ -17,20 +18,20 @@ int_to_mod = {
     1 << 12    : ["SO",     "SpunOut"],
     1 << 13    : ["AP",   "Autopilot"],
     1 << 14    : ["PF",     "Perfect"],
-    1 << 15    : ["K4",        "Key4"],
-    1 << 16    : ["K5",        "Key5"],
-    1 << 17    : ["K6",        "Key6"],
-    1 << 18    : ["K7",        "Key7"],
-    1 << 19    : ["K8",        "Key8"],
+    1 << 15    : ["4K",        "Key4"],
+    1 << 16    : ["5K",        "Key5"],
+    1 << 17    : ["6K",        "Key6"],
+    1 << 18    : ["7K",        "Key7"],
+    1 << 19    : ["8K",        "Key8"],
     1 << 20    : ["FI",      "FadeIn"],
     1 << 21    : ["RD",      "Random"],
     1 << 22    : ["CN",      "Cinema"],
     1 << 23    : ["TP",      "Target"],
-    1 << 24    : ["K9",        "Key9"],
+    1 << 24    : ["9K",        "Key9"],
     1 << 25    : ["CO",     "KeyCoop"],
-    1 << 26    : ["K1",        "Key1"],
-    1 << 27    : ["K3",        "Key3"],
-    1 << 28    : ["K2",        "Key2"],
+    1 << 26    : ["1K",        "Key1"],
+    1 << 27    : ["3K",        "Key3"],
+    1 << 28    : ["2K",        "Key2"],
     1 << 29    : ["V2",     "ScoreV2"],
     1 << 30    : ["MR",      "Mirror"]
 }
@@ -99,7 +100,7 @@ class ModCombination(BaseModel):
         return mod.value
 
     def short_name(self):
-        """
+        r"""
         The acronym-ized names of the component mods.
 
         Returns
@@ -133,7 +134,7 @@ class ModCombination(BaseModel):
         return "".join(mod.short_name() for mod in component_mods)
 
     def long_name(self):
-        """
+        r"""
         The spelled out names of the component mods.
 
         Returns
@@ -190,7 +191,7 @@ class ModCombination(BaseModel):
         return bool(self.value & other.value)
 
     def decompose(self, clean=False):
-        """
+        r"""
         Decomposes this mod into its base component mods, which are
         :class:`~.ModCombination`\s with a ``value`` of a power of two.
 
