@@ -6,17 +6,69 @@ from typing import Optional, TypeVar, Generic, Any, List, Union, Dict
 from dataclasses import dataclass
 
 from ossapi.mod import Mod
-from ossapi.enums import (UserAccountHistory, ProfileBanner, UserBadge, Country,
-    Cover, UserGroup, UserMonthlyPlaycount, UserPage, UserReplaysWatchedCount,
-    UserAchievement, UserProfileCustomization, RankHistory, Kudosu, PlayStyles,
-    ProfilePage, GameMode, RankStatus, Failtimes, Covers, Hype, Availability,
-    Nominations, Statistics, Grade, Weight, MessageType, KudosuAction,
-    KudosuGiver, KudosuPost, EventType, EventAchivement, EventUser,
-    EventBeatmap, BeatmapsetApproval, EventBeatmapset, KudosuVote,
-    BeatmapsetEventType, UserRelationType, UserLevel, UserGradeCounts,
-    GithubUser, ChangelogSearch, ForumTopicType, ForumPostBody, ForumTopicSort,
-    ChannelType, ReviewsConfig, NewsSearch, Nomination, RankHighest, RoomType,
-    RoomCategory, MatchEventType, ScoringType, TeamType)
+from ossapi.enums import (
+    UserAccountHistory,
+    ProfileBanner,
+    UserBadge,
+    Country,
+    Cover,
+    UserGroup,
+    UserMonthlyPlaycount,
+    UserPage,
+    UserReplaysWatchedCount,
+    UserAchievement,
+    UserProfileCustomization,
+    RankHistory,
+    Kudosu,
+    PlayStyles,
+    ProfilePage,
+    GameMode,
+    RankStatus,
+    Failtimes,
+    Covers,
+    Hype,
+    Availability,
+    Nominations,
+    Statistics,
+    Grade,
+    Weight,
+    MessageType,
+    KudosuAction,
+    KudosuGiver,
+    KudosuPost,
+    EventType,
+    EventAchivement,
+    EventUser,
+    EventBeatmap,
+    BeatmapsetApproval,
+    EventBeatmapset,
+    KudosuVote,
+    BeatmapsetEventType,
+    UserRelationType,
+    UserLevel,
+    UserGradeCounts,
+    GithubUser,
+    ChangelogSearch,
+    ForumTopicType,
+    ForumPostBody,
+    ForumTopicSort,
+    ChannelType,
+    ReviewsConfig,
+    NewsSearch,
+    Nomination,
+    RankHighest,
+    RoomType,
+    RoomCategory,
+    MatchEventType,
+    ScoringType,
+    TeamType,
+    Variant,
+    ForumPollText,
+    ForumPollTitle,
+    BeatmapPackUserCompletionData,
+    RoomPlaylistItemStats,
+    RoomDifficultyRange,
+)
 from ossapi.utils import Datetime, Model, BaseModel, Field
 
 T = TypeVar("T")
@@ -55,6 +107,7 @@ is, not that the api actually lets any type be returned there.
 # We want cursors to also be instantiatable manually (eg `Cursor(page=199)`),
 # so make `_data` optional and also allow arbitrary `kwargs`.
 
+
 class Cursor(BaseModel):
     def __init__(self, _data=None, **kwargs):
         super().__init__()
@@ -72,21 +125,24 @@ class Cursor(BaseModel):
     def __eq__(self, other):
         return self.__dict__ == other.__dict__
 
+
 # if there are no more results, a null cursor is returned instead.
 # So always let the cursor be nullable to catch this. It's the user's
 # responsibility to check for a null cursor to see if there are any more
 # results.
 CursorT = Optional[Cursor]
+CursorStringT = Optional[str]
+
 
 class UserCompact(Model):
     """
     https://osu.ppy.sh/docs/index.html#usercompact
     """
+
     # required fields
     # ---------------
     avatar_url: str
     country_code: str
-    default_group: str
     id: int
     is_active: bool
     is_bot: bool
@@ -102,27 +158,20 @@ class UserCompact(Model):
     # ---------------
     account_history: Optional[List[UserAccountHistory]]
     active_tournament_banner: Optional[ProfileBanner]
+    active_tournament_banners: Optional[List[ProfileBanner]]
     badges: Optional[List[UserBadge]]
     beatmap_playcounts_count: Optional[int]
     blocks: Optional[UserRelation]
     country: Optional[Country]
     cover: Optional[Cover]
+    default_group: Optional[str]
     favourite_beatmapset_count: Optional[int]
-    # undocumented
     follow_user_mapping: Optional[List[int]]
     follower_count: Optional[int]
     friends: Optional[List[UserRelation]]
     graveyard_beatmapset_count: Optional[int]
     groups: Optional[List[UserGroup]]
-    # undocumented
     guest_beatmapset_count: Optional[int]
-    is_admin: Optional[bool]
-    is_bng: Optional[bool]
-    is_full_bn: Optional[bool]
-    is_gmt: Optional[bool]
-    is_limited_bn: Optional[bool]
-    is_moderator: Optional[bool]
-    is_nat: Optional[bool]
     is_restricted: Optional[bool]
     is_silenced: Optional[bool]
     loved_beatmapset_count: Optional[int]
@@ -130,7 +179,11 @@ class UserCompact(Model):
     mapping_follower_count: Optional[int]
     monthly_playcounts: Optional[List[UserMonthlyPlaycount]]
     page: Optional[UserPage]
+    pending_beatmapset_count: Optional[int]
     previous_usernames: Optional[List[str]]
+    # deprecated, replaced by rank_history
+    rankHistory: Optional[RankHistory]
+    rank_history: Optional[RankHistory]
     # deprecated, replaced by ranked_beatmapset_count
     ranked_and_approved_beatmapset_count: Optional[int]
     ranked_beatmapset_count: Optional[int]
@@ -143,17 +196,15 @@ class UserCompact(Model):
     support_level: Optional[int]
     # deprecated, replaced by pending_beatmapset_count
     unranked_beatmapset_count: Optional[int]
-    pending_beatmapset_count: Optional[int]
     unread_pm_count: Optional[int]
     user_achievements: Optional[List[UserAchievement]]
     user_preferences: Optional[UserProfileCustomization]
-    rank_history: Optional[RankHistory]
-    # deprecated, replaced by rank_history
-    rankHistory: Optional[RankHistory]
+    session_verified: Optional[bool]
 
     def expand(self) -> User:
         return self._fk_user(self.id)
 
+
 class User(UserCompact):
     comments_count: int
     cover_url: str
@@ -170,6 +221,8 @@ class User(UserCompact):
     playstyle: Optional[PlayStyles]
     post_count: int
     profile_order: List[ProfilePage]
+    profile_hue: Optional[int]
+    daily_challenge_user_stats: DailyChallengeUserStats
     title: Optional[str]
     title_url: Optional[str]
     twitter: Optional[str]
@@ -197,7 +250,7 @@ class BeatmapCompact(Model):
 
     # optional fields
     # ---------------
-    _beatmapset: Optional[BeatmapsetCompact] = Field(name="beatmapset")
+    _beatmapset: Field(name="beatmapset", type=Optional[BeatmapsetCompact])
     checksum: Optional[str]
     failtimes: Optional[Failtimes]
     max_combo: Optional[int]
@@ -209,8 +262,8 @@ class BeatmapCompact(Model):
         return self._fk_user(self.user_id)
 
     def beatmapset(self) -> Union[Beatmapset, BeatmapsetCompact]:
-        return self._fk_beatmapset(self.beatmapset_id,
-            existing=self._beatmapset)
+        return self._fk_beatmapset(self.beatmapset_id, existing=self._beatmapset)
+
 
 class Beatmap(BeatmapCompact):
     total_length: int
@@ -238,24 +291,24 @@ class Beatmap(BeatmapCompact):
     # beatmapset owner.
     # This is optional as a workaround until
     # https://github.com/ppy/osu-web/issues/9784 is resolved.
-    owner: Optional[UserCompact] = Field(name="user")
+    owner: Field(name="user", type=Optional[UserCompact])
 
     # overridden fields
     # -----------------
-    _beatmapset: Optional[Beatmapset] = Field(name="beatmapset")
+    _beatmapset: Field(name="beatmapset", type=Optional[Beatmapset])
 
     def expand(self) -> Beatmap:
         return self
 
     def beatmapset(self) -> Beatmapset:
-        return self._fk_beatmapset(self.beatmapset_id,
-            existing=self._beatmapset)
+        return self._fk_beatmapset(self.beatmapset_id, existing=self._beatmapset)
 
 
 class BeatmapsetCompact(Model):
     """
     https://osu.ppy.sh/docs/index.html#beatmapsetcompact
     """
+
     # required fields
     # ---------------
     artist: str
@@ -264,17 +317,17 @@ class BeatmapsetCompact(Model):
     creator: str
     favourite_count: int
     id: int
+    nsfw: bool
+    offset: int
     play_count: int
     preview_url: str
     source: str
     status: RankStatus
+    spotlight: bool
     title: str
     title_unicode: str
     user_id: int
     video: bool
-    nsfw: bool
-    offset: int
-    spotlight: bool
     # documented as being in `Beatmapset` only, but returned by
     # `api.beatmapset_events` which uses a `BeatmapsetCompact`.
     hype: Optional[Hype]
@@ -283,6 +336,7 @@ class BeatmapsetCompact(Model):
     # ---------------
     beatmaps: Optional[List[Beatmap]]
     converts: Optional[Any]
+    current_nominations: Optional[List[Nomination]]
     current_user_attributes: Optional[Any]
     description: Optional[Any]
     discussions: Optional[Any]
@@ -291,12 +345,12 @@ class BeatmapsetCompact(Model):
     has_favourited: Optional[bool]
     language: Optional[Any]
     nominations: Optional[Any]
+    pack_tags: Optional[List[str]]
     ratings: Optional[Any]
     recent_favourites: Optional[Any]
     related_users: Optional[Any]
-    _user: Optional[UserCompact] = Field(name="user")
-    # undocumented
     track_id: Optional[int]
+    _user: Field(name="user", type=Optional[UserCompact])
 
     def expand(self) -> Beatmapset:
         return self._fk_beatmapset(self.id)
@@ -304,10 +358,12 @@ class BeatmapsetCompact(Model):
     def user(self) -> Union[UserCompact, User]:
         return self._fk_user(self.user_id, existing=self._user)
 
+
 class Beatmapset(BeatmapsetCompact):
     availability: Availability
     bpm: float
     can_be_hyped: bool
+    deleted_at: Optional[Datetime]
     discussion_enabled: bool
     discussion_locked: bool
     is_scoreable: bool
@@ -319,23 +375,19 @@ class Beatmapset(BeatmapsetCompact):
     storyboard: bool
     submitted_date: Optional[Datetime]
     tags: str
-    current_nominations: Optional[List[Nomination]]
-    deleted_at: Optional[Datetime]
-    pack_tags: List[str]
 
     def expand(self) -> Beatmapset:
         return self
 
+
 # undocumented, but defined here to avoid a forward reference in Score.
 class ScoreMatchInfo(Model):
     slot: int
     team: str
-    pass_: bool = Field(name="pass")
+    pass_: Field(name="pass", type=bool)
 
-class Score(Model):
-    """
-    https://osu.ppy.sh/docs/index.html#score
-    """
+
+class _LegacyScore(Model):
     # can be null for match scores, eg the scores
     # in https://osu.ppy.sh/community/matches/97947404
     id: Optional[int]
@@ -355,21 +407,79 @@ class Score(Model):
     replay: bool
     passed: bool
     current_user_attributes: Any
+    beatmap: Optional[Beatmap]
+    beatmapset: Optional[BeatmapsetCompact]
+    rank_country: Optional[int]
+    rank_global: Optional[int]
+    weight: Optional[Weight]
+    _user: Field(name="user", type=Optional[UserCompact])
+    match: Optional[ScoreMatchInfo]
+    type: str
+
+
+class Score(Model):
+    """
+    https://osu.ppy.sh/docs/index.html#score with x-api-version >= 20220705
+    """
+
+    id: Optional[int]
+    best_id: Optional[int]
+    user_id: int
+    accuracy: float
+    max_combo: int
+    statistics: Statistics
+    pp: Optional[float]
+    rank: Grade
+
+    passed: bool
+    current_user_attributes: Any
+    classic_total_score: int
+    processed: bool
+    replay: bool
+    maximum_statistics: Any  # TODO property typing
+    mods: List[NonLegacyMod]
+    ruleset_id: int
+    started_at: Optional[Datetime]
+    ended_at: Optional[Datetime]
+    ranked: bool
+    preserve: bool
+    beatmap_id: int
+    build_id: Optional[int]
+    has_replay: bool
+    is_perfect_combo: bool
+    total_score: int
+    total_score_without_mods: Optional[int]
+
+    legacy_perfect: bool
+    legacy_score_id: Optional[int]
+    legacy_total_score: int
 
     beatmap: Optional[Beatmap]
     beatmapset: Optional[BeatmapsetCompact]
     rank_country: Optional[int]
     rank_global: Optional[int]
     weight: Optional[Weight]
-    _user: Optional[UserCompact] = Field(name="user")
+    _user: Field(name="user", type=Optional[UserCompact])
     match: Optional[ScoreMatchInfo]
     type: str
 
-    @classmethod
-    def preprocess_data(cls, data):
+    @staticmethod
+    def override_attributes(data, api):
+        if api.api_version < 20220705:
+            return _LegacyScore
+        # there are rare cases where a legacy score is returned even when using a
+        # modern api version. Legacy matches is the only exception I'm aware of currently.
+        #
+        # check a few attributes to be reasonably certain that we are in this case,
+        # and then switch to _LegacyScore.
+        if "mode" in data and "created_at" in data and "legacy_perfect" not in data:
+            return _LegacyScore
+
+    @staticmethod
+    def preprocess_data(data, api):
         # scores from matches (api.match) return perfect as an int instead of a
         # bool (same as api v1). Convert to a bool here.
-        if isinstance(data["perfect"], int):
+        if "perfect" in data and isinstance(data["perfect"], int):
             data["perfect"] = bool(data["perfect"])
         return data
 
@@ -379,16 +489,19 @@ class Score(Model):
     def download(self):
         return self._api.download_score(self.mode, self.id)
 
+
 class BeatmapUserScore(Model):
     position: int
     score: Score
 
+
 class BeatmapUserScores(Model):
     scores: List[Score]
 
+
 class BeatmapScores(Model):
     scores: List[Score]
-    userScore: Optional[BeatmapUserScore]
+    user_score: Field(name="userScore", type=Optional[BeatmapUserScore])
 
 
 class CommentableMeta(Model):
@@ -401,9 +514,11 @@ class CommentableMeta(Model):
     owner_title: Optional[str]
     current_user_attributes: Optional[CommentableMetaCurrentUserAttributes]
 
+
 class CommentableMetaCurrentUserAttributes(Model):
     can_new_comment_reason: Optional[str]
 
+
 class Comment(Model):
     # null for deleted commentables, eg on /comments/3.
     commentable_id: Optional[int]
@@ -431,21 +546,23 @@ class Comment(Model):
     def edited_by(self) -> Optional[User]:
         return self._fk_user(self.edited_by_id)
 
+
 class CommentBundle(Model):
     commentable_meta: List[CommentableMeta]
     comments: List[Comment]
+    cursor: CursorT
     has_more: bool
     has_more_id: Optional[int]
     included_comments: List[Comment]
     pinned_comments: Optional[List[Comment]]
+    # TODO this should be type CommentSort
     sort: str
     top_level_count: Optional[int]
     total: Optional[int]
     user_follow: bool
     user_votes: List[int]
     users: List[UserCompact]
-    # undocumented
-    cursor: CursorT
+
 
 class ForumPost(Model):
     created_at: Datetime
@@ -464,6 +581,7 @@ class ForumPost(Model):
     def edited_by(self) -> Optional[User]:
         return self._fk_user(self.edited_by_id)
 
+
 class ForumTopic(Model):
     created_at: Datetime
     deleted_at: Optional[Datetime]
@@ -477,32 +595,55 @@ class ForumTopic(Model):
     type: ForumTopicType
     updated_at: Datetime
     user_id: int
-    poll: Any
+    poll: Optional[ForumPollModel]
 
     def user(self) -> User:
         return self._fk_user(self.user_id)
 
+
+class ForumPollModel(Model):
+    allow_vote_change: bool
+    ended_at: Optional[Datetime]
+    hide_incomplete_results: bool
+    last_vote_at: Optional[Datetime]
+    max_votes: int
+    options: List[ForumPollOption]
+    started_at: Datetime
+    title: ForumPollTitle
+    total_vote_count: int
+
+
+class ForumPollOption(Model):
+    id: int
+    text: ForumPollText
+    vote_count: Optional[int]
+
+
 class ForumTopicAndPosts(Model):
     cursor: CursorT
     search: ForumTopicSearch
     posts: List[ForumPost]
     topic: ForumTopic
-    cursor_string: Optional[str]
+    cursor_string: CursorStringT
+
 
 class CreateForumTopicResponse(Model):
     post: ForumPost
     topic: ForumTopic
 
+
 class ForumTopicSearch(Model):
     sort: Optional[ForumTopicSort]
     limit: Optional[int]
     start: Optional[int]
     end: Optional[int]
 
+
 class SearchResult(Generic[T], Model):
     data: List[T]
     total: int
 
+
 class WikiPage(Model):
     layout: str
     locale: str
@@ -513,9 +654,11 @@ class WikiPage(Model):
     title: str
     available_locales: List[str]
 
+
 class Search(Model):
-    users: Optional[SearchResult[UserCompact]] = Field(name="user")
-    wiki_pages: Optional[SearchResult[WikiPage]] = Field(name="wiki_page")
+    users: Field(name="user", type=Optional[SearchResult[UserCompact]])
+    wiki_pages: Field(name="wiki_page", type=Optional[SearchResult[WikiPage]])
+
 
 class Spotlight(Model):
     end_date: Datetime
@@ -526,23 +669,34 @@ class Spotlight(Model):
     start_date: Datetime
     type: str
 
+
 class Spotlights(Model):
     spotlights: List[Spotlight]
 
+
 # return-value wrapper for https://osu.ppy.sh/docs/index.html#get-users.
 class Users(Model):
     users: List[UserCompact]
 
+
 # return-value wrapper for https://osu.ppy.sh/docs/index.html#get-beatmaps.
 class Beatmaps(Model):
     beatmaps: List[Beatmap]
 
+
+class BeatmapPacks(Model):
+    cursor: CursorT
+    cursor_string: CursorStringT
+    beatmap_packs: List[BeatmapPack]
+
+
 class Rankings(Model):
     beatmapsets: Optional[List[Beatmapset]]
     cursor: CursorT
-    ranking: List[UserStatistics]
+    ranking: Union[List[UserStatistics], List[CountryStatistics]]
     spotlight: Optional[Spotlight]
-    total: int
+    total: Optional[int]
+
 
 class BeatmapsetDiscussionPost(Model):
     id: int
@@ -556,7 +710,7 @@ class BeatmapsetDiscussionPost(Model):
     updated_at: Datetime
     deleted_at: Optional[Datetime]
 
-    def user(self) -> user:
+    def user(self) -> User:
         return self._fk_user(self.user_id)
 
     def last_editor(self) -> Optional[User]:
@@ -565,6 +719,7 @@ class BeatmapsetDiscussionPost(Model):
     def deleted_by(self) -> Optional[User]:
         return self._fk_user(self.deleted_by_id)
 
+
 class BeatmapsetDiscussion(Model):
     id: int
     beatmapset_id: int
@@ -579,19 +734,15 @@ class BeatmapsetDiscussion(Model):
     can_be_resolved: bool
     can_grant_kudosu: bool
     created_at: Datetime
-    # documented as non-optional, api.beatmapset_events() might give a null
-    # response for this? but very rarely. need to find a repro case
     current_user_attributes: Any
     updated_at: Datetime
     deleted_at: Optional[Datetime]
-    # similarly as for current_user_attributes, in the past this has been null
-    # but can't find a repro case
     last_post_at: Datetime
     kudosu_denied: bool
     starting_post: Optional[BeatmapsetDiscussionPost]
     posts: Optional[List[BeatmapsetDiscussionPost]]
-    _beatmap: Optional[BeatmapCompact] = Field(name="beatmap")
-    _beatmapset: Optional[BeatmapsetCompact] = Field(name="beatmapset")
+    _beatmap: Field(name="beatmap", type=Optional[BeatmapCompact])
+    _beatmapset: Field(name="beatmapset", type=Optional[BeatmapsetCompact])
 
     def user(self) -> User:
         return self._fk_user(self.user_id)
@@ -600,12 +751,12 @@ class BeatmapsetDiscussion(Model):
         return self._fk_user(self.deleted_by_id)
 
     def beatmapset(self) -> Union[Beatmapset, BeatmapsetCompact]:
-        return self._fk_beatmapset(self.beatmapset_id,
-            existing=self._beatmapset)
+        return self._fk_beatmapset(self.beatmapset_id, existing=self._beatmapset)
 
     def beatmap(self) -> Union[Optional[Beatmap], BeatmapCompact]:
         return self._fk_beatmap(self.beatmap_id, existing=self._beatmap)
 
+
 class BeatmapsetDiscussionVote(Model):
     id: int
     score: int
@@ -613,11 +764,14 @@ class BeatmapsetDiscussionVote(Model):
     beatmapset_discussion_id: int
     created_at: Datetime
     updated_at: Datetime
-    cursor_string: Optional[str]
+    # TODO is this field ever actually returned? not documented and can't find
+    # a repro case.
+    cursor_string: CursorStringT
 
     def user(self):
         return self._fk_user(self.user_id)
 
+
 class KudosuHistory(Model):
     id: int
     action: KudosuAction
@@ -631,9 +785,10 @@ class KudosuHistory(Model):
     # see https://github.com/ppy/osu-web/issues/7549
     details: Any
 
+
 class BeatmapPlaycount(Model):
     beatmap_id: int
-    _beatmap: Optional[BeatmapCompact] = Field(name="beatmap")
+    _beatmap: Field(name="beatmap", type=Optional[BeatmapCompact])
     beatmapset: Optional[BeatmapsetCompact]
     count: int
 
@@ -644,8 +799,8 @@ class BeatmapPlaycount(Model):
 # we use this class to determine which event dataclass to instantiate and
 # return, based on the value of the ``type`` parameter.
 class _Event(Model):
-    @classmethod
-    def override_class(cls, data):
+    @staticmethod
+    def override_attributes(data, api):
         mapping = {
             EventType.ACHIEVEMENT: AchievementEvent,
             EventType.BEATMAP_PLAYCOUNT: BeatmapPlaycountEvent,
@@ -659,45 +814,54 @@ class _Event(Model):
             EventType.USER_SUPPORT_FIRST: UserSupportFirstEvent,
             EventType.USER_SUPPORT_AGAIN: UserSupportAgainEvent,
             EventType.USER_SUPPORT_GIFT: UserSupportGiftEvent,
-            EventType.USERNAME_CHANGE: UsernameChangeEvent
+            EventType.USERNAME_CHANGE: UsernameChangeEvent,
         }
         type_ = EventType(data["type"])
         return mapping[type_]
 
+
 class Event(Model):
     created_at: Datetime
     createdAt: Datetime
     id: int
     type: EventType
 
+
 class AchievementEvent(Event):
     achievement: EventAchivement
     user: EventUser
 
+
 class BeatmapPlaycountEvent(Event):
     beatmap: EventBeatmap
     count: int
 
+
 class BeatmapsetApproveEvent(Event):
     approval: BeatmapsetApproval
     beatmapset: EventBeatmapset
     user: EventUser
 
+
 class BeatmapsetDeleteEvent(Event):
     beatmapset: EventBeatmapset
 
+
 class BeatmapsetReviveEvent(Event):
     beatmapset: EventBeatmapset
     user: EventUser
 
+
 class BeatmapsetUpdateEvent(Event):
     beatmapset: EventBeatmapset
     user: EventUser
 
+
 class BeatmapsetUploadEvent(Event):
     beatmapset: EventBeatmapset
     user: EventUser
 
+
 class RankEvent(Event):
     scoreRank: str
     rank: int
@@ -705,23 +869,29 @@ class RankEvent(Event):
     beatmap: EventBeatmap
     user: EventUser
 
+
 class RankLostEvent(Event):
     mode: GameMode
     beatmap: EventBeatmap
     user: EventUser
 
+
 class UserSupportFirstEvent(Event):
     user: EventUser
 
+
 class UserSupportAgainEvent(Event):
     user: EventUser
 
+
 class UserSupportGiftEvent(Event):
-    beatmap: EventBeatmap
+    user: EventUser
+
 
 class UsernameChangeEvent(Event):
     user: EventUser
 
+
 class Build(Model):
     created_at: Datetime
     display_version: str
@@ -731,11 +901,14 @@ class Build(Model):
     version: Optional[str]
     changelog_entries: Optional[List[ChangelogEntry]]
     versions: Optional[Versions]
+    youtube_id: Optional[str]
+
 
 class Versions(Model):
     next: Optional[Build]
     previous: Optional[Build]
 
+
 class UpdateStream(Model):
     display_name: Optional[str]
     id: int
@@ -744,6 +917,7 @@ class UpdateStream(Model):
     latest_build: Optional[Build]
     user_count: Optional[int]
 
+
 class ChangelogEntry(Model):
     category: str
     created_at: Optional[Datetime]
@@ -759,18 +933,21 @@ class ChangelogEntry(Model):
     url: Optional[str]
     github_user: GithubUser
 
+
 class ChangelogListing(Model):
     builds: List[Build]
     search: ChangelogSearch
     streams: List[UpdateStream]
 
+
 class MultiplayerScores(Model):
-    cursor_string: str
+    cursor_string: CursorStringT
     params: str
     scores: List[MultiplayerScore]
     total: Optional[int]
     user_score: Optional[MultiplayerScore]
 
+
 class MultiplayerScore(Model):
     id: int
     user_id: int
@@ -790,17 +967,20 @@ class MultiplayerScore(Model):
     def beatmap(self):
         return self._fk_beatmap(self.beatmap_id)
 
+
 class MultiplayerScoresAround(Model):
     higher: List[MultiplayerScore]
     lower: List[MultiplayerScore]
 
+
 class NewsListing(Model):
     cursor: CursorT
-    cursor_string: str
+    cursor_string: CursorStringT
     news_posts: List[NewsPost]
     news_sidebar: NewsSidebar
     search: NewsSearch
 
+
 class NewsPost(Model):
     author: str
     edit_url: str
@@ -814,26 +994,32 @@ class NewsPost(Model):
     navigation: Optional[NewsNavigation]
     preview: Optional[str]
 
+
 class NewsNavigation(Model):
     newer: Optional[NewsPost]
     older: Optional[NewsPost]
 
+
 class NewsSidebar(Model):
     current_year: int
     news_posts: List[NewsPost]
     years: List[int]
 
+
 class SeasonalBackgrounds(Model):
     ends_at: Datetime
     backgrounds: List[SeasonalBackground]
 
+
 class SeasonalBackground(Model):
     url: str
     user: UserCompact
 
+
 class DifficultyAttributes(Model):
     attributes: BeatmapDifficultyAttributes
 
+
 class BeatmapDifficultyAttributes(Model):
     max_combo: int
     star_rating: float
@@ -851,8 +1037,8 @@ class BeatmapDifficultyAttributes(Model):
     stamina_difficulty: Optional[float]
     rhythm_difficulty: Optional[float]
     colour_difficulty: Optional[float]
-    approach_raty: Optional[float]
-    great_hit_windoy: Optional[float]
+    approach_rate: Optional[float]
+    great_hit_window: Optional[float]
 
     # ctb attributes
     approach_rate: Optional[float]
@@ -862,6 +1048,26 @@ class BeatmapDifficultyAttributes(Model):
     score_multiplier: Optional[float]
 
 
+class Events(Model):
+    cursor: CursorT
+    cursor_string: CursorStringT
+    events: Field(type=List[_Event])
+
+
+class BeatmapPack(Model):
+    author: str
+    date: Datetime
+    name: str
+    no_diff_reduction: bool
+    # marked as nonnull on docs
+    ruleset_id: Optional[int]
+    tag: str
+    url: str
+
+    # optional attributes
+    beatmapsets: Optional[List[Beatmapset]]
+    user_completion_data: Optional[BeatmapPackUserCompletionData]
+
 
 # ================
 # Parameter Models
@@ -869,6 +1075,7 @@ class BeatmapDifficultyAttributes(Model):
 
 # models which aren't used for serialization, but passed to OssapiV2 methods.
 
+
 @dataclass
 class ForumPoll:
     options: List[str]
@@ -881,11 +1088,11 @@ class ForumPoll:
     vote_change: bool = False
 
 
-
 # ===================
 # Undocumented Models
 # ===================
 
+
 class BeatmapsetSearchResult(Model):
     beatmapsets: List[Beatmapset]
     cursor: CursorT
@@ -895,6 +1102,7 @@ class BeatmapsetSearchResult(Model):
     search: Any
     cursor_string: Optional[str]
 
+
 class BeatmapsetDiscussions(Model):
     beatmaps: List[Beatmap]
     cursor: CursorT
@@ -905,11 +1113,13 @@ class BeatmapsetDiscussions(Model):
     cursor_string: Optional[str]
     beatmapsets: List[Beatmapset]
 
+
 class BeatmapsetDiscussionReview(Model):
     # https://github.com/ppy/osu-web/blob/master/app/Libraries/BeatmapsetDis
     # cussionReview.php
     max_blocks: int
 
+
 class BeatmapsetDiscussionPosts(Model):
     beatmapsets: List[BeatmapsetCompact]
     discussions: List[BeatmapsetDiscussion]
@@ -918,6 +1128,7 @@ class BeatmapsetDiscussionPosts(Model):
     users: List[UserCompact]
     cursor_string: Optional[str]
 
+
 class BeatmapsetDiscussionVotes(Model):
     cursor: CursorT
     discussions: List[BeatmapsetDiscussion]
@@ -925,14 +1136,17 @@ class BeatmapsetDiscussionVotes(Model):
     users: List[UserCompact]
     cursor_string: Optional[str]
 
+
 class BeatmapsetEventComment(Model):
     beatmap_discussion_id: int
     beatmap_discussion_post_id: int
 
+
 class BeatmapsetEventCommentNoPost(Model):
     beatmap_discussion_id: int
     beatmap_discussion_post_id: None
 
+
 class BeatmapsetEventCommentNone(Model):
     beatmap_discussion_id: None
     beatmap_discussion_post_id: None
@@ -942,35 +1156,44 @@ class BeatmapsetEventCommentChange(Generic[S], BeatmapsetEventCommentNone):
     old: S
     new: S
 
+
 class BeatmapsetEventCommentLovedRemoval(BeatmapsetEventCommentNone):
     reason: str
 
+
 class BeatmapsetEventCommentKudosuChange(BeatmapsetEventCommentNoPost):
     new_vote: KudosuVote
     votes: List[KudosuVote]
 
+
 class BeatmapsetEventCommentKudosuRecalculate(BeatmapsetEventCommentNoPost):
     new_vote: Optional[KudosuVote]
 
+
 class BeatmapsetEventCommentOwnerChange(BeatmapsetEventCommentNone):
     beatmap_id: int
     beatmap_version: str
     new_user_id: int
     new_user_username: str
 
+
 class BeatmapsetEventCommentNominate(Model):
     # for some reason this comment type doesn't have the normal
     # beatmap_discussion_id and beatmap_discussion_post_id attributes (they're
     # not even null, just missing).
     modes: List[GameMode]
 
+
 class BeatmapsetEventCommentWithNominators(BeatmapsetEventCommentNoPost):
+    beatmap_ids: Optional[List[int]]
     nominator_ids: Optional[List[int]]
 
+
 class BeatmapsetEventCommentWithSourceUser(BeatmapsetEventCommentNoPost):
     source_user_id: int
     source_user_username: str
 
+
 class BeatmapsetEvent(Model):
     # https://github.com/ppy/osu-web/blob/master/app/Models/BeatmapsetEvent.php
     #
@@ -979,14 +1202,15 @@ class BeatmapsetEvent(Model):
 
     id: int
     type: BeatmapsetEventType
-    comment: str
+    comment: Any
     created_at: Datetime
 
     user_id: Optional[int]
     beatmapset: Optional[BeatmapsetCompact]
     discussion: Optional[BeatmapsetDiscussion]
 
-    def override_types(self):
+    @staticmethod
+    def override_attributes(data, api):
         mapping = {
             BeatmapsetEventType.BEATMAP_OWNER_CHANGE: BeatmapsetEventCommentOwnerChange,
             BeatmapsetEventType.DISCUSSION_DELETE: BeatmapsetEventCommentNoPost,
@@ -1022,30 +1246,31 @@ class BeatmapsetEvent(Model):
             BeatmapsetEventType.REMOVE_FROM_LOVED: BeatmapsetEventCommentLovedRemoval,
             BeatmapsetEventType.NSFW_TOGGLE: BeatmapsetEventCommentChange[bool],
         }
-        type_ = BeatmapsetEventType(self.type)
+        type_ = BeatmapsetEventType(data["type"])
         return {"comment": mapping[type_]}
 
     def user(self) -> Optional[User]:
         return self._fk_user(self.user_id)
 
+
 class ChatChannel(Model):
     channel_id: int
     description: Optional[str]
     icon: Optional[str]
-    # documented as non-optional (to see that it can be null, pm tillerino)
     moderated: Optional[bool]
     name: str
     type: ChannelType
     uuid: Optional[str]
+    message_length_limit: int
 
     # optional fields
     # ---------------
-    first_message_id: Optional[int]
     last_message_id: Optional[int]
     last_read_id: Optional[int]
     recent_messages: Optional[List[ChatMessage]]
     users: Optional[List[int]]
 
+
 class ChatMessage(Model):
     channel_id: int
     content: str
@@ -1057,6 +1282,16 @@ class ChatMessage(Model):
     # TODO enumify, example value: "plain"
     type: str
 
+
+class CountryStatistics(Model):
+    code: str
+    active_users: int
+    play_count: int
+    ranked_score: int
+    performance: int
+    country: Country
+
+
 class CreatePMResponse(Model):
     message: ChatMessage
     new_channel_id: int
@@ -1067,6 +1302,7 @@ class CreatePMResponse(Model):
     # documented but not present in response
     presence: Optional[List[ChatChannel]]
 
+
 class ModdingHistoryEventsBundle(Model):
     # https://github.com/ppy/osu-web/blob/master/app/Libraries/ModdingHistor
     # yEventsBundle.php#L84
@@ -1074,6 +1310,7 @@ class ModdingHistoryEventsBundle(Model):
     reviewsConfig: BeatmapsetDiscussionReview
     users: List[UserCompact]
 
+
 class UserRelation(Model):
     # undocumented (and not a class on osu-web)
     # https://github.com/ppy/osu-web/blob/master/app/Transformers/UserRelatio
@@ -1090,30 +1327,41 @@ class UserRelation(Model):
         return self._fk_user(self.target_id, existing=self.target)
 
 
-class UserStatistics(Model):
-    level: UserLevel
-    pp: float
-    ranked_score: int
-    hit_accuracy: float
-    play_count: int
-    play_time: int
-    total_score: int
-    total_hits: int
-    maximum_combo: int
-    replays_watched_by_others: int
-    is_ranked: bool
-    grade_counts: UserGradeCounts
+class StatisticsVariant(Model):
+    mode: GameMode
+    variant: Variant
     country_rank: Optional[int]
     global_rank: Optional[int]
-    rank: Optional[Any]
-    user: Optional[UserCompact]
-    variants: Optional[Any]
-    global_rank_exp: Optional[float]
-    pp_exp: float
+    pp: float
+
+
+class UserStatistics(Model):
     count_100: int
     count_300: int
     count_50: int
     count_miss: int
+    country_rank: Optional[int]
+    grade_counts: UserGradeCounts
+    hit_accuracy: float
+    is_ranked: bool
+    level: UserLevel
+    maximum_combo: int
+    play_count: int
+    play_time: Optional[int]
+    pp: Optional[float]
+    pp_exp: float
+    global_rank: Optional[int]
+    global_rank_exp: Optional[float]
+    # deprecated, replaced by global_rank and country_rank
+    rank: Optional[Any]
+    ranked_score: int
+    rank_change_since_30_days: Optional[int]
+    replays_watched_by_others: int
+    total_hits: int
+    total_score: int
+    user: Optional[UserCompact]
+    variants: Optional[List[StatisticsVariant]]
+
 
 class UserStatisticsRulesets(Model):
     # undocumented
@@ -1124,10 +1372,12 @@ class UserStatisticsRulesets(Model):
     fruits: Optional[UserStatistics]
     mania: Optional[UserStatistics]
 
+
 class RoomPlaylistItemMod(Model):
     acronym: str
     settings: Dict[str, Any]
 
+
 class RoomPlaylistItem(Model):
     id: int
     room_id: int
@@ -1143,7 +1393,8 @@ class RoomPlaylistItem(Model):
     played_at: Optional[Datetime]
     beatmap: BeatmapCompact
 
-class Room(Model):
+
+class _Room1(Model):
     id: int
     name: str
     category: RoomCategory
@@ -1162,6 +1413,37 @@ class Room(Model):
     playlist: List[RoomPlaylistItem]
     recent_participants: List[UserCompact]
 
+
+class Room(Model):
+    id: int
+    name: str
+    category: RoomCategory
+    type: RoomType
+    user_id: int
+    starts_at: Datetime
+    ends_at: Datetime
+    max_attempts: Optional[int]
+    participant_count: int
+    channel_id: int
+    active: bool
+    has_password: bool
+    queue_mode: str
+    auto_skip: bool
+    host: UserCompact
+    playlist: List[RoomPlaylistItem]
+
+    # new from _Room1
+    playlist_item_stats: RoomPlaylistItemStats
+    current_playlist_item: Optional[RoomPlaylistItem]
+    difficulty_range: RoomDifficultyRange
+    recent_participants: List[UserCompact]
+
+    @staticmethod
+    def override_attributes(data, api):
+        if api.api_version < 20220217:
+            return _Room1
+
+
 class RoomLeaderboardScore(Model):
     accuracy: float
     attempts: int
@@ -1172,13 +1454,16 @@ class RoomLeaderboardScore(Model):
     user_id: int
     user: UserCompact
 
+
 class RoomLeaderboardUserScore(RoomLeaderboardScore):
     position: int
 
+
 class RoomLeaderboard(Model):
     leaderboard: List[RoomLeaderboardScore]
     user_score: RoomLeaderboardUserScore
 
+
 class Match(Model):
     id: int
     start_time: Datetime
@@ -1186,11 +1471,13 @@ class Match(Model):
     end_time: Optional[Datetime]
     name: str
 
+
 class Matches(Model):
     matches: List[Match]
     cursor: CursorT
     params: Any
-    cursor_string: str
+    cursor_string: CursorStringT
+
 
 class MatchGame(Model):
     id: int
@@ -1202,15 +1489,20 @@ class MatchGame(Model):
     scoring_type: ScoringType
     team_type: TeamType
     mods: List[Mod]
-    beatmap: BeatmapCompact
+    # null for deleted beatmaps,
+    # e.g. https://osu.ppy.sh/community/matches/103721175.
+    # TODO doesn't match docs
+    beatmap: Optional[BeatmapCompact]
     beatmap_id: int
     scores: List[Score]
 
+
 class MatchEventDetail(Model):
     type: MatchEventType
     # seems to only be used for MatchEventType.OTHER
     text: Optional[str]
 
+
 class MatchEvent(Model):
     id: int
     detail: MatchEventDetail
@@ -1219,6 +1511,7 @@ class MatchEvent(Model):
     user_id: Optional[int]
     game: Optional[MatchGame]
 
+
 class MatchResponse(Model):
     match: Match
     events: List[MatchEvent]
@@ -1226,3 +1519,21 @@ class MatchResponse(Model):
     first_event_id: int
     latest_event_id: int
     current_game_id: Optional[int]
+
+
+class DailyChallengeUserStats(Model):
+    daily_streak_best: int
+    daily_streak_current: int
+    last_update: Datetime
+    last_weekly_streak: Datetime
+    playcount: int
+    top_10p_placements: int
+    top_50p_placements: int
+    user_id: int
+    weekly_streak_best: int
+    weekly_streak_current: int
+
+
+class NonLegacyMod(Model):
+    acronym: str
+    settings: Any
