@@ -1,9 +1,10 @@
 from enum import Enum, IntFlag
 from datetime import datetime, timezone
-from typing import Any
+from typing import Any, Union
 from dataclasses import dataclass
 
-from typing_utils import issubtype
+from typing_utils import issubtype, get_origin, get_args
+
 
 def is_high_model_type(type_):
     """
@@ -14,6 +15,7 @@ def is_high_model_type(type_):
     """
     return is_model_type(type_) and not is_base_model_type(type_)
 
+
 def is_model_type(type_):
     """
     Whether ``type_`` is a subclass of ``Model``.
@@ -22,6 +24,7 @@ def is_model_type(type_):
         return False
     return issubclass(type_, Model)
 
+
 def is_base_model_type(type_):
     """
     Whether ``type_`` is a subclass of ``BaseModel``.
@@ -32,9 +35,20 @@ def is_base_model_type(type_):
 
 
 class Field:
-    def __init__(self, *, name=None):
+    def __init__(self, *, name=None, type=None):
         self.name = name
 
+        # We use annotations for two distinct purposes: deserialization, and
+        # user-facing type hints. If the deserialize type does not match the
+        # runtime type, these two annotations need to be different.
+        #
+        # For instance, events should deserialize as _Event, but have a runtime
+        # type of Event.
+        #
+        # This field allows setting the runtime type via annotations and the
+        # deserialize type via passing this field.
+        self.type = type
+
 
 class _Model:
     """
@@ -42,7 +56,9 @@ class _Model:
     its own members and cleanup after instantion, subclass ``BaseModel``
     instead.
     """
-    def override_types(self):
+
+    @staticmethod
+    def override_attributes(data, api):
         """
         Sometimes, the types of attributes in models depends on the value of
         other fields in that model. By overriding this method, models can return
@@ -51,29 +67,13 @@ class _Model:
         instead.
 
         This method should return a mapping of ``attribute_name`` to
-        ``intended_type``.
-        """
-        return {}
-
-    @classmethod
-    def override_class(cls, _data):
-        """
-        This method addressess a shortcoming in ``override_types`` in order to
-        achieve full coverage of the intended feature of overriding types.
-
-        The model that we want to override types for may be at the very top of
-        the hierarchy, meaning we can't go any higher and find a model for which
-        we can override ``override_types`` to customize this class' type.
-
-        A possible solution for this is to create a wrapper class one step above
-        it; however, this is both dirty and may not work (I haven't actually
-        tried it). So this method provides a way for a model to override its
-        *own* type (ie class) at run-time.
+        ``intended_type``. You can also return a class to use exactly the
+        attributes of that class.
         """
         return None
 
-    @classmethod
-    def preprocess_data(cls, data):
+    @staticmethod
+    def preprocess_data(data, api):
         """
         A hook that allows model classes to modify data from the api before it
         is used to instantiate the model.
@@ -89,6 +89,7 @@ class _Model:
         """
         return data
 
+
 class ModelMeta(type):
     def __new__(cls, name, bases, dct):
         model = super().__new__(cls, name, bases, dct)
@@ -104,16 +105,29 @@ class ModelMeta(type):
                 continue
             setattr(model, name, None)
 
-        return dataclass(model)
+        return model
+
 
 class Model(_Model, metaclass=ModelMeta):
     """
     A dataclass-style model. Provides an ``_api`` attribute.
     """
-    # This is the ``OssapiV2`` instance that loaded this model.
-    # can't annotate with OssapiV2 or we get a circular import error, this is
-    # good enough.
-    _api: Any
+
+    def __init__(self, **kwargs):
+        self._ossapi_data = kwargs
+
+    def __getattribute__(self, key):
+        data = super().__getattribute__("_ossapi_data")
+        if key == "_ossapi_data":
+            return data
+        if key in data:
+            return data[key]
+        return super().__getattribute__(key)
+
+    def __setattr__(self, key, value):
+        if key == "_ossapi_data":
+            return super().__setattr__(key, value)
+        self._ossapi_data[key] = value
 
     def _foreign_key(self, fk, func, existing):
         if existing:
@@ -134,6 +148,26 @@ class Model(_Model, metaclass=ModelMeta):
         func = lambda: self._api.beatmapset(beatmapset_id)
         return self._foreign_key(beatmapset_id, func, existing)
 
+    def __str__(self):
+        # don't print internal values
+        blacklisted_keys = ["_api"]
+        items = [
+            f"{k}={v!r}"
+            for k, v in self._ossapi_data.items()
+            if k not in blacklisted_keys
+        ]
+        return "{}({})".format(type(self).__name__, ", ".join(items))
+
+    def __eq__(self, other):
+        if not isinstance(other, type(self)):
+            return NotImplemented
+        return all(
+            getattr(self, key) == getattr(other, key) for key in self._ossapi_data
+        )
+
+    __repr__ = __str__
+
+
 class BaseModel(_Model):
     """
     A model which promises to take care of its own members and cleanup, after we
@@ -150,11 +184,14 @@ class BaseModel(_Model):
     for instance). We don't need or want to do anything else with an enum after
     instantiating it, hence it's defined as a base type.
     """
+
     pass
 
+
 class EnumModel(BaseModel, Enum):
     pass
 
+
 class IntFlagModel(BaseModel, IntFlag):
     pass
 
@@ -164,7 +201,8 @@ class Datetime(datetime, BaseModel):
     Our replacement for the ``datetime`` object that deals with the various
     datetime formats the api returns.
     """
-    def __new__(cls, value): # pylint: disable=signature-differs
+
+    def __new__(cls, value):
         if value is None:
             raise ValueError("cannot instantiate a Datetime with a null value")
         # the api returns a bunch of different timestamps: two ISO 8601
@@ -204,10 +242,10 @@ class Datetime(datetime, BaseModel):
         return True
 
 
-
 # typing utils
 # ------------
 
+
 def is_optional(type_):
     """
     Whether ``type(None)`` is a valid instance of ``type_``. eg,
@@ -218,18 +256,46 @@ def is_optional(type_):
     ``Union[Any, str]`` is a valid type not equal to ``Any`` (in python), but
     representing the same set of types.
     """
+
+    # issubtype is expensive as it requires normalizing the types. It's possible
+    # we could improve performance there, but there is low hanging fruit for
+    # is_optional in particular: the vast majority of calls are on very simple
+    # types, such as primitive types (int, str, etc), model types (UserCompact),
+    # or the simplest form of optional type (Optional[int]). For these cases,
+    # we can do much better with faster checks.
+    #
+    # It's rare that complicated types such as Union[Union[int, None], str] come
+    # up which  require normalization. However, we'll keep the issubtype check
+    # as the "ground truth" fallback for when the optimizations fail.
+    #
+    # This method is used in tight deserialization loops, so it's important to
+    # keep it inexpensive.
+
+    # optimization for common case of primitive types
+    if type_ in [int, float, str, bool]:
+        return False
+
+    # optimization for common case of simple optional - Optional[int] ie
+    # Union[int, NoneType].
+    if get_origin(type_) is Union:
+        if get_args(type_)[1] is type(None):
+            return True
+
     return issubtype(type(None), type_) and not issubtype(Any, type_)
 
+
 def is_primitive_type(type_):
     if not isinstance(type_, type):
         return False
     return type_ in [int, float, str, bool]
 
-def is_compatible_type(value, type_):
-    # make an exception for an integer being instantiated as a float. In
-    # the json we receive, eg ``pp`` can have a value of ``15833``, which is
-    # interpreted as an int by our json parser even though ``pp`` is a
-    # float.
+
+def convert_primitive_type(value, type_):
+    # In the json we receive, eg ``pp`` can have a value of ``15833``, which is
+    # interpreted as an int by our json parser even though ``pp`` is a float.
+    # Convert back to the correct typing here.
+    # This is important as some consumers may rely on float-specific methods
+    # (`#is_integer`)
     if type_ is float and isinstance(value, int):
-        return True
-    return isinstance(value, type_)
+        return float(value)
+    return value
