@@ -1,7 +1,6 @@
 from unittest import TestCase
 
-from ossapi import (User, BeatmapsetCompact, UserCompact, GameMode,
-    BeatmapCompact)
+from ossapi import User, BeatmapsetCompact, UserCompact, GameMode, BeatmapCompact
 
 from tests import api_v2 as api
 
@@ -10,17 +9,17 @@ class TestMethodTypeConversion(TestCase):
     def test_beatmap_as_parameter(self):
         # make sure we can pass the full model in place of an id
         beatmap = api.beatmap(221777)
-        self.assertEqual(api.beatmap_scores(beatmap),
-            api.beatmap_scores(beatmap.id))
+        assert api.beatmap_scores(beatmap) == api.beatmap_scores(beatmap.id)
 
     def test_room_as_parameter(self):
         room1 = api.room(257524)
         room2 = api.room(room1)
         self.assertEqual(room1, room2)
 
+
 class TestExpandableModels(TestCase):
     def test_expand_user(self):
-        user = api.search(query="tybug2").users.data[0]
+        user = api.search(query="tybug").users.data[0]
         # `statistics` is only available on User models, so make sure it's not
         # present before expanding and is present afterwards
         self.assertIsNone(user.statistics)
@@ -45,6 +44,7 @@ class TestExpandableModels(TestCase):
     # TODO add test_expand_beatmap when I find a good endpoint that returns
     # BeatmapCompacts and not full Beatmaps.
 
+
 class TestFollowingForeignKeys(TestCase):
     def test_beatmap_fks(self):
         beatmap = api.beatmap(221777)
@@ -66,7 +66,7 @@ class TestFollowingForeignKeys(TestCase):
         self.assertEqual(user.id, 4693052)
 
     def test_score_fks(self):
-        score = api.score(GameMode.OSU, 3685255338)
+        score = api.score_mode(GameMode.OSU, 3685255338)
 
         user = score.user()
         self.assertIsInstance(user, UserCompact)
